# SPDX-FileCopyrightText: UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

# class Tagger(_Tagger, frozen=True, tag=str.lower):
# files: Set[str] = field(default_factory=set)


# class Names(Files, frozen=True):
#     names: List[str] = field(default_factory=list)
#
#
# class Fnmatches(Files, frozen=True):
#     patterns: List[str] = field(default_factory=list)



# SPDX-FileCopyrightText: UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import re
from typing import Set

from msgspec import field

from ._base import _Tagger

_EXTENSIONS = {}
_TAGS = {}


class Tagger(_Tagger, frozen=True, tag="file.extension"):
    extensions: Set[str] = field(default_factory=set)


def register(tagger: Tagger):
    for tag in tagger.tags:
        _EXTENSIONS.setdefault(tag, set())
        _EXTENSIONS[tag] |= tagger.extensions

    for extension in tagger.extensions:
        _TAGS.setdefault(extension, set())
        _TAGS[extension] |= tagger.tags


# multi-line regex

# def find_tags(files: Set[str]) -> Set[str]:
#     pat_tokens = '|'.join('^(?P<%s>%s)$' % (extension, r".+\." + extension) for extension in _TAGS)
#     re_extensions = re.compile(pat_tokens, flags=re.MULTILINE)
#     tags_found = {}
#     for match in re_extensions.finditer('\n'.join(files)):
#         for tag in _TAGS[match.lastgroup]:
#             tags_found.setdefault(tag, set())
#             tags_found[tag].add(match.group())
#     return tags_found

# line-by-line regex


def find_tags(files: Set[str]) -> Set[str]:
    pat_tokens = "|".join(
        "^(?P<%s>%s)$" % (extension, r".+\." + extension) for extension in _TAGS
    )
    re_extensions = re.compile(pat_tokens)
    tags_found = {}
    for file in files:
        match = re_extensions.match(file)
        if match:
            for tag in _TAGS.get(match.lastgroup, set()):
                tags_found.setdefault(tag, set())
                tags_found[tag].add(match.group())
    return tags_found


# fnmatch filter

# def find_tags(files: Set[str]) -> Set[str]:
#     tags_found = {}
#     for extension in _TAGS:
#         files_found = set(fnmatch.filter(files, f"*.{extension}"))
#         if files_found:
#             for tag in _TAGS[extension]:
#                 tags_found.setdefault(tag, set())
#                 tags_found[tag] |= files_found
#     return tags_found

# match the extension

# def find_tags(files: Set[str]) -> Set[str]:
#     tags_found = {}
#     for file in files:
#         extension = Path(file).suffix.lower().replace(".", "")
#         if extension in _TAGS:
#             for tag in _TAGS[extension]:
#                 tags_found.setdefault(tag, set())
#                 tags_found[tag].add(file)

#     return tags_found



# SPDX-FileCopyrightText: UL Research Institutes
# SPDX-License-Identifier: Apache-2.0


import pytest

from taggie.main import get_parser


@pytest.mark.parametrize(
    "argv,expected,succeeds,error,code,out,err",
    (
        ([], dict(), True, None, None, None, None),
        # -h/--help
        (["-h"], dict(), False, SystemExit, 0, None, None),
        (["--help"], dict(), False, SystemExit, 0, None, None),
        # -V/--version
        (["-V"], dict(version=True), False, SystemExit, 0, "taggie 0.0.0", None),
        (["--version"], dict(version=True), False, SystemExit, 0, "taggie 0.0.0", None),
        # -g/--group
        (["-g", "language"], dict(groups=["language"]), True, None, None, None, None),
        (
            ["--group", "language"],
            dict(groups=["language"]),
            True,
            None,
            None,
            None,
            None,
        ),
        (["-g"], None, False, SystemExit, 2, None, None),
        (["--group"], None, False, SystemExit, 2, None, None),
        # -s/--sort
        # default choice is most
        ([], dict(sort_by="most"), True, None, None, None, None),
        ([], dict(sort_by="most"), True, None, None, None, None),
        # accepts choices
        (["-s", "most"], dict(sort_by="most"), True, None, None, None, None),
        (["--sort", "most"], dict(sort_by="most"), True, None, None, None, None),
        (["-s", "name"], dict(sort_by="name"), True, None, None, None, None),
        (["--sort", "name"], dict(sort_by="name"), True, None, None, None, None),
        # rejects invalid choice
        (["-s", "else"], None, False, SystemExit, 2, None, "invalid choice"),
        (["--sort", "else"], None, False, SystemExit, 2, None, "invalid choice"),
        # rejects no argument
        (["-s"], None, False, SystemExit, 2, None, "expected one argument"),
        (["--sort"], None, False, SystemExit, 2, None, "expected one argument"),
    ),
)
def test_parse_args(argv, expected, succeeds, error, code, out, err, capsys):
    parser = get_parser()
    if succeeds:
        args = parser.parse_args(argv)
        for key, value in expected.items():
            assert getattr(args, key) == value
    else:
        if error == SystemExit:
            with pytest.raises(error, check=lambda e: e.code == code):
                args = parser.parse_args(argv)
            captured = capsys.readouterr()
            if out:
                assert out in captured.out
            if err:
                assert err in captured.err
        else:
            with pytest.raises(error):
                args = parser.parse_args(argv)



# SPDX-FileCopyrightText: UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import argparse
import json
import os
from importlib import import_module
from pathlib import Path

# from .taggers.files import Fnmatches, Extensions, Names
from typing import List, Set

import msgspec
from tabulate import tabulate

from ._version import __version__

PACKAGE_DIR = Path(__file__).parent.absolute()

DEFAULTS_DIR = PACKAGE_DIR / "defaults"

TAGGER_PACKAGE_DIRS = [PACKAGE_DIR / "taggers"]

TAGGER_PACKAGES = {
    path.stem.replace("_", "."): import_module(f".{path.stem}", "taggie.taggers")
    for tagger_package_dir in TAGGER_PACKAGE_DIRS
    for path in tagger_package_dir.glob("*.py")
    if not path.stem.startswith("_")
}

TAGGER_TYPES = [tagger.Tagger for tagger in TAGGER_PACKAGES.values()]


def get_tag_directory(tag_path: Path) -> Set[str]:
    tag_files = set()
    for root, _, files in os.walk(tag_path):
        root_path = Path(root)
        for file in files:
            if file.endswith(".yaml"):
                tag_files.add(str(root_path / file))
    return tag_files


def get_tag_files(tag_files: List[str]) -> Set[str]:
    all_tag_files = set()
    for tag_file in tag_files:
        tag_path = Path(tag_file)
        if tag_path.is_dir():
            all_tag_files |= get_tag_directory(tag_path)
        else:
            all_tag_files.add(tag_file)
    return all_tag_files


def get_all_files(path: Path) -> Set[str]:
    all_files = set()
    for root, _, files in os.walk(path):
        root_path = Path(root)
        relative_path = root_path.relative_to(path)
        for file in files:
            all_files.add(str(relative_path / file))
    return all_files


class SetEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, set):
            return list(obj)
        return json.JSONEncoder.default(self, obj)


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument("-t", "--tag-file", dest="tag_files", action="append")
    parser.add_argument("-g", "--group", dest="groups", action="append")
    parser.add_argument(
        "-s", "--sort", dest="sort_by", choices=["most", "name"], default="most"
    )
    parser.add_argument(
        "-V", "--version", action="version", version=f"taggie {__version__}"
    )
    return parser


def main():
    parser = get_parser()
    args = parser.parse_args()

    if not args.tag_files:
        args.tag_files = [DEFAULTS_DIR.absolute()]

    base_path = Path.cwd()

    # print(TAGGER_PACKAGES)
    # print(TAGGER_TYPES)

    all_tag_files = get_tag_files(args.tag_files)
    # print(all_tag_files)

    all_files = get_all_files(base_path)
    # print(all_files)

    # map_file_tags = {}
    # map_tag_files = {}

    for tag_file in all_tag_files:
        taggers = msgspec.yaml.decode(
            open(tag_file).read(), type=List[tuple(TAGGER_TYPES)]
        )
        for tagger in taggers:
            if args.groups and not any(fi in tagger.groups for fi in args.groups):
                continue

            tagger_type = (
                Path(tagger.__class__.__module__)
                .suffix.lower()
                .replace(".", "")
                .replace("_", ".")
            )
            TAGGER_PACKAGES[tagger_type].register(tagger)

    all_found_tags = {}
    for tagger_type, tagger_package in TAGGER_PACKAGES.items():
        found_tags = tagger_package.find_tags(all_files)
        for tag, files in found_tags.items():
            all_found_tags.setdefault(tag, set())
            all_found_tags[tag] |= files

    # print(all_found_tags)

    table = []
    for tag, files in found_tags.items():
        table.append([len(files), tag, list(files)[0]])

    if args.sort_by == "most":
        table = sorted(table, key=lambda k: k[0], reverse=True)
    elif args.sort_by == "name":
        table = sorted(table, key=lambda k: k[1])

    if table:
        print(
            tabulate(
                table,
                tablefmt="plain",
                headers=["TOTAL", "NAME", "EXAMPLE"],
                colalign=["left", "left", "left"],
            )
        )



# SPDX-FileCopyrightText: UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

from typing import Set

from msgspec import Struct, field


class _Tagger(Struct, frozen=True, tag=str.lower):
    tags: Set[str] = field(default_factory=set)
    groups: Set[str] = field(default_factory=set)
